package ORG.ebiss.sara.interpreter;

import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.parser.BinaryOpExpression;

public class InitialBindingLoader {

  public static void execute(Interpreter interpreter, Interactor interactor) {
    try {
      NameSpace global = interpreter.getGlobalNameSpace();
      SpdValue root = SpdValue._object = new SpdValue();

      root.defineMethod("copy", new PrimitiveMethod(0) {
	public Code execute(Interpreter interpreter, SpdValue receiver) {
	  return interpreter.popFrameAndReturn(receiver.copy());
	}
      });

      root.defineMethod("new", new Method() {
	public Code apply(Interpreter interpreter,
			  SpdValue container, SpdValue receiver)
	  throws EvaluationException
	{
	  SpdValue result = ((SpdValue)receiver).freshCopy();
	  Value init = result.read("init");
	  if (init == null) return interpreter.popFrameAndReturn(result);
	  if (! (init instanceof ApplicableValue)) {
	    throw new EvaluationException("init must be an applicable value");
	  }
	  int[] arguments = interpreter.getArgumentsVector();
	  Code returnPoint = interpreter.popFrameAndReturn(result);
	  return ((ApplicableValue)init).
	    apply(interpreter, arguments, -1, returnPoint);
	}

	public Object getPrintableRepresentation(PrintingContext context) {
	  return "(...)";
	}

	public boolean checkArity(int arity) { return true; }
      });

      root.defineMethod("join", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  Value argument = interpreter.readArgument(0);
	  if (! (argument instanceof SpdValue)) return null;
	  try {
	    receiver.join((SpdValue)argument);
	  } catch (InvalidOperationException uoe) {
	    throw new EvaluationException("failed to join objects");
	  }
	  return interpreter.popFrameAndReturn(receiver);
	}
      });

      root.defineMethod("hashCode", new PrimitiveMethod(0) {
	public Code execute(Interpreter interpreter, SpdValue receiver) {
	  return interpreter.
	    popFrameAndReturn(new IntegerValue(receiver.hashCode()));
	}
      });

      root.defineMethod("equals", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    Value argument = interpreter.readArgument(0);
	    return interpreter.
	      popFrameAndReturn(receiver.binaryOp(BinaryOpExpression.EQUAL,
						  argument));
	  } catch (InvalidOperationException uoe) {
	    throw new EvaluationException("unsupported operation");
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("domain error");
	  }
	}
      });

      root.defineMethod("getPrintableRepresentation", new PrimitiveMethod(0) {
	public Code execute(Interpreter interpreter, SpdValue receiver) {
	  return interpreter.
	    popFrameAndReturn(new StringValue(receiver.
					      getPrintableRepresentation()));
	}
      });

      root.defineMethod("inherits", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver) {
	  Value argument = interpreter.readArgument(0);
	  return interpreter.
	    popFrameAndReturn(argument instanceof SpdValue &&
			      receiver.inherits((SpdValue)argument) ?
			      BooleanValue.TRUE : BooleanValue.FALSE);
	}
      });

      root.defineMethod("isDefined", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    return interpreter.
	      popFrameAndReturn(receiver.read(name) != null ?
				BooleanValue.TRUE : BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("the argument must be " +
					  "a string or a symbol");
	  }
	}
      });

      global.define("Object", root);

      SpdValue system = new SpdValue();

      system.defineVariable("interactor", new JavaValue(interactor));
      system.defineVariable("interpreter", new JavaValue(interpreter));
      system.defineVariable(SpdValue.inheritModifier, "parent", root);
      system.defineMethod("getTimeStamp", new PrimitiveMethod(0) {
	public Code execute(Interpreter interpreter, SpdValue receiver) {
	  String	s;

	  StringBuffer result = new StringBuffer();
	  Calendar now = Calendar.getInstance();

	  s = "0" + (now.get(Calendar.MONTH) + (1 - Calendar.JANUARY));
	  s = s.substring(s.length() - 2);
	  result.append(s);

	  s = "0" + now.get(Calendar.DAY_OF_MONTH);
	  s = s.substring(s.length() - 2);
	  result.append(s);

	  s = "0" + now.get(Calendar.HOUR_OF_DAY);
	  s = s.substring(s.length() - 2);
	  result.append(s);

	  s = "0" + now.get(Calendar.MINUTE);
	  s = s.substring(s.length() - 2);
	  result.append(s);

	  s = "0" + now.get(Calendar.SECOND);
	  s = s.substring(s.length() - 2);
	  result.append(s);

	  return interpreter.
	    popFrameAndReturn(new StringValue(result.toString()));
	}
      });
      global.define("System", system);

      global.define("isString", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof StringValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isSymbol", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof SymbolValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isInteger", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof IntegerValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isFloat", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof DoubleValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isBoolean", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof BooleanValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isNull", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value == NULL) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isObject", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof SpdValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isList", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof ListValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isVector", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  interpreter.
	    writeRegister(resultSlot, (value instanceof VectorValue) ?
			  BooleanValue.TRUE : BooleanValue.FALSE);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("isHashMap", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  BooleanValue bool = BooleanValue.FALSE;
	  if (value instanceof JavaValue) {
	    if ( ((JavaValue)value).getJavaValue() instanceof HashMap) {
	      bool = BooleanValue.TRUE;
	    }
	  }
	  interpreter.
	    writeRegister(resultSlot,bool);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });


      global.define("isDefined", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  Value value = interpreter.readRegister(arguments[0]);
	  try {
	    interpreter.
	      writeRegister(resultSlot,
			    interpreter.isDefinedVariable(value.stringValue())
			    ? BooleanValue.TRUE : BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("the argument must be a string " +
					  "or a symbol");
	  }
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("toPrintableString", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  try {
	    Value value = interpreter.readRegister(arguments[0]);
	    StringValue result =
	      new StringValue(value.getPrintableRepresentation());
	    interpreter.writeRegister(resultSlot, result);
	    return returnPoint;
	  } catch (ClassCastException cce) {
	    throw new EvaluationException("object is expected");
	  }
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      global.define("java", new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint)
	  throws EvaluationException
	{
	  if (arguments.length != 1) {
	    throw new EvaluationException("invalid arguments");
	  }
	  String name = null;
	  try {
	    name = interpreter.readRegister(arguments[0]).stringValue();
	    Value result = new JavaValue(Class.forName(name));
	    interpreter.writeRegister(resultSlot, result);
	    return returnPoint;
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("class name is expected");
	  } catch (ClassNotFoundException cnfe) {
	    throw new EvaluationException("Class not found: \"" +
					  name + "\"");
	  }
	}

	public boolean checkArity(int arity) { return arity == 1; }
      });

      root.defineMethod("importSlot", new PrimitiveMethod(3) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    Value object = interpreter.readArgument(1);
	    String slotName = interpreter.readArgument(2).stringValue();
	    if (! (object instanceof SpdValue)) {
	      throw new EvaluationException("the 2nd argument of " +
					    "importSlot must be a Object");
	    }
	    receiver.importSlot(name, (SpdValue)object, slotName);
	    return interpreter.popFrameAndReturn(Value.UNSPECIFIED);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("A string is expected");
	  } catch (InvalidOperationException uoe) {
	    throw new EvaluationException("Unsupported operation");
	  }
	}
      });

      root.defineMethod("readSlot", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	     throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    Value result = receiver.read(name);
	    return interpreter.
	      popFrameAndReturn(result != null ? result : Value.NULL);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected " +
					  "as 1st argument");
	  }
	}
      });

      root.defineMethod("writeSlot", new PrimitiveMethod(2) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    Value value = interpreter.readArgument(1);
	    boolean result = receiver.write(name, value);
	    return interpreter.popFrameAndReturn(result ? BooleanValue.TRUE
							: BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  } catch (InvalidOperationException uoe) {
	    throw new EvaluationException("invalid value");
	  }
	}
      });

      root.defineMethod("slots", new PrimitiveMethod(0) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  return interpreter.
	    popFrameAndReturn(new JavaValue(receiver.elements()));
	}
      });

      root.defineMethod("isObservable", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    Slot slot = receiver.lookupSlot(name);
	    if (slot == null) {
	      throw new EvaluationException("Unknown slot: " + name);
	    }
	    return interpreter.
	      popFrameAndReturn(slot instanceof ObservableSlot
				? BooleanValue.TRUE
				: BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("isAction", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    boolean result = receiver.isAction(name);
	    return interpreter.popFrameAndReturn(result ? BooleanValue.TRUE
							: BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("isVisible", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	     throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    boolean result = receiver.isVisible(name);
	    return interpreter.popFrameAndReturn(result ? BooleanValue.TRUE
						        : BooleanValue.FALSE);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("setVisible", new PrimitiveMethod(2) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	     throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    boolean value = interpreter.readArgument(1).booleanValue();
	    receiver.setVisible(name, value);
	    return interpreter.popFrameAndReturn(Value.UNSPECIFIED);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("getLabel", new PrimitiveMethod(1) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	     throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    String label = receiver.getLabel(name);
	    return interpreter.popFrameAndReturn(new StringValue(label));
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("setLabel", new PrimitiveMethod(2) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	     throws EvaluationException
	{
	  try {
	    String name = interpreter.readArgument(0).stringValue();
	    String value = interpreter.readArgument(1).stringValue();
	    receiver.setLabel(name, value);
	    return interpreter.popFrameAndReturn(Value.UNSPECIFIED);
	  } catch (CoerceFailedException cfe) {
	    throw new EvaluationException("slot name is expected as " +
					  "1st argument");
	  }
	}
      });

      root.defineMethod("getChangeList", new PrimitiveMethod(3) {
	public Code execute(Interpreter interpreter, SpdValue receiver)
	  throws EvaluationException
	{
	  try {
	    Array addList =
	      (Array)(interpreter.readArgument(0).javaValue());
	    Array removeList =
	      (Array)(interpreter.readArgument(1).javaValue());
	    Array modifiedList =
	      (Array)(interpreter.readArgument(2).javaValue());
	    receiver.getChangeList(addList, removeList, modifiedList);
	    return interpreter.popFrameAndReturn(Value.UNSPECIFIED);
	  } catch (InvalidOperationException ioe) {
	    throw new EvaluationException("the value of property must be an " +
					  "atomic value or a flat container");
	  } catch (ClassCastException cce) {
	    cce.printStackTrace();
	    throw new EvaluationException("three vectors are expected " +
					  "as the arguments");
	  }
	}
      });
    } catch (InvalidOperationException uoe) {
      uoe.printStackTrace();
      throw new RuntimeException();
    }
  }
}
