package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.parser.Node;
import ORG.ebiss.sara.parser.Parser;
import ORG.ebiss.sara.parser.ParseException;
import ORG.ebiss.sara.parser.BinaryOpExpression;
import ORG.ebiss.sara.parser.LocalizedCharStream;

public class Interactor {
  public BufferedReader		in;
  public PrintWriter		out;
  public PrintWriter		err;
  public java.awt.Frame		interactionFrame;
  protected OrderedMap		_commands = new OrderedMap(new LessString());
  protected java.util.Stack	_contextStack = new java.util.Stack();
  protected boolean		_leaveInteractor;
  protected Interpreter		_interpreter;

  public interface Command {
    public void execute(String argument) throws EvaluationException;
  }

  public Interactor(final Interpreter interpreter, String consoleServerPath)
       throws IOException, UnknownHostException
  {
    _interpreter = interpreter;
    _contextStack.push(_interpreter.saveContext());

    FileInputStream ttyin = new FileInputStream("/dev/tty");
    FileOutputStream ttyout = new FileOutputStream("/dev/tty");
    System.setIn(ttyin);
    System.setOut(new PrintStream(ttyout));
    System.setErr(new PrintStream(ttyout));
    in = new BufferedReader(new InputStreamReader(ttyin));
    out = new PrintWriter(ttyout);
    err = out;

    _commands.add("gui", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.equals("on")) {
	  if (interactionFrame == null) {
	    interactionFrame = createInteractionFrame();
	  }
	  interactionFrame.setVisible(true);
	} else if (argument.equals("off")) {
	  interactionFrame.setVisible(false);
	} else {
	  out.println("Usage: gui (on|off)");
	}
      }
    });

    _commands.add("quit", new Command() {
      public void execute(String argument) throws EvaluationException {
	quit();
      }
    });

    _commands.add("resume", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (_contextStack.size() > 1) {
	  _contextStack.pop();
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  _interpreter.resume();
	  _contextStack.pop();
	  _contextStack.push(_interpreter.saveContext());
	} else {
	  _interpreter.loadInitialContext();
	}
      }
    });

    _commands.add("top", new Command() {
      public void execute(String argument) throws EvaluationException {
	_interpreter.loadInitialContext();
	_contextStack.removeAllElements();
	_contextStack.push(_interpreter.saveContext());
      }
    });

    _commands.add("continue", new Command() {
      public void execute(String argument) throws EvaluationException {
	_interpreter.restoreContext((Context)_contextStack.peek());
	_interpreter.resume();
	_contextStack.pop();
	_contextStack.push(_interpreter.saveContext());
      }
    });

    _commands.add("finish", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: finish <expression>");
	  return;
	}
	try {
	  _contextStack.push(_contextStack.peek());
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  Parser parser = new Parser(new LocalizedCharStream(argument));
	  Node expression = parser.Expression();
	  Value result = _interpreter.evaluate(expression);
	  _contextStack.pop();
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  _interpreter.finishProcedure(result);
	  _contextStack.pop();
	  _contextStack.push(_interpreter.saveContext());
	} catch (ParseException pe) {
	  out.println(pe.getMessage());
	}
      }
    });

    _commands.add("istep", new Command() {
      public void execute(String argument) throws EvaluationException {
	_interpreter.restoreContext((Context)_contextStack.peek());
	Code currentCode = _interpreter.currentCode();
	if (currentCode == null) return;
	out.println(currentCode);
	_interpreter.istep();
	_contextStack.pop();
	_contextStack.push(_interpreter.saveContext());
      }
    });

    _commands.add("where", new Command() {
      public void execute(String argument) throws EvaluationException {
	_interpreter.restoreContext((Context)_contextStack.peek());
	out.print(_interpreter.dumpCallStack());
	out.flush();
      }
    });

    _commands.add("setbp", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  if (_interpreter.breakPoints == null) {
	    out.println("There are no break points");
	  } else {
	    out.println("Current break points:");
	    for (int j = 0; j < _interpreter.breakPoints.length; ++j) {
	      out.print(j);
	      out.print(": ");
	      out.println(_interpreter.breakPoints[j]);
	    }
	  }
	  return;
	}
	try {
	  int address = Integer.parseInt(argument);
	  if (_interpreter.breakPoints == null) {
	    _interpreter.breakPoints = new int[1];
	    _interpreter.breakPoints[0] = address;
	  } else {
	    int[] bp = _interpreter.breakPoints;
	    int[] newbp = new int[bp.length + 1];
	    System.arraycopy(bp, 0, newbp, 0, bp.length);
	    _interpreter.breakPoints = newbp;
	    _interpreter.breakPoints[bp.length] = address;
	  }
	} catch (NumberFormatException nfe) {
	  out.println("Usage: setbp <address>");
	}
      }
    });

    _commands.add("unsetbp", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  _interpreter.breakPoints = null;
	  return;
	}
	try {
	  int index = Integer.parseInt(argument);
	  int[] bp = _interpreter.breakPoints;
	  if (bp == null || bp.length <= index || index < 0) {
	    out.println("invalid index");
	  } else {
	    if (bp.length == 1) {
	      _interpreter.breakPoints = null;
	    } else {
	      int[] newbp = _interpreter.breakPoints = new int[bp.length - 1];
	      System.arraycopy(bp, 0, newbp, 0, index);
	      if (index < newbp.length) {
		System.arraycopy(bp, index + 1,
				 newbp, index, newbp.length - index);
	      }
	    }
	  }
	} catch (NumberFormatException nfe) {
	  out.println("Usage: unsetbp <index>");
	}
      }
    });

    _commands.add("trace", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.equals("on")) {
	  _interpreter.trace = true;
	} else if (argument.equals("off")) {
	  _interpreter.trace = false;
	} else {
	  out.println("Usage: trace (on|off)");
	}
      }
    });

    _commands.add("print", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: print <expression>");
	  return;
	}
	try {
	  _contextStack.push(_contextStack.peek());
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  Parser parser = new Parser(new LocalizedCharStream(argument));
	  Node expression = parser.Expression();
	  Value result = _interpreter.evaluate(expression);
	  out.println(result.getPrintableRepresentation());
	  _contextStack.pop();
	} catch (ParseException pe) {
	  out.println(pe.getMessage());
	}
      }
    });

    _commands.add("read", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: read <register number>");
	  return;
	}
	try {
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  int register = Integer.parseInt(argument);
	  Value result = _interpreter.readRegister(register);
	  out.println(result.getPrintableRepresentation());
	} catch (NumberFormatException nfe) {
	  out.println("Usage: read <register number>");
	}
      }
    });

    _commands.add("write", new Command() {
      public void execute(String argument) throws EvaluationException {
	try {
	  _contextStack.push(_contextStack.peek());
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  Parser parser = new Parser(new LocalizedCharStream(argument));
	  ORG.ebiss.sara.parser.Literal registerNode = parser.Literal();
	  if (! (registerNode.value instanceof Integer)) throw new Exception();
	  int register = ((Integer)registerNode.value).intValue();
	  Node expression = parser.Expression();
	  Value value = _interpreter.evaluate(expression);
	  _contextStack.pop();
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  _interpreter.writeRegister(register, value);
	  out.println(value.getPrintableRepresentation());
	} catch (EvaluationException ee) {
	  throw ee;
	} catch (Exception others) {
	  out.println("Usage: write <register number> <expression>");
	  return;
	}
      }
    });

    _commands.add("do", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: do <statement>");
	  return;
	}
	try {
	  _interpreter.restoreContext((Context)_contextStack.peek());
	  Parser parser = new Parser(new LocalizedCharStream(argument));
	  Node statement = parser.Program();
	  _interpreter.execute(statement);
	  _contextStack.pop();
	  _contextStack.push(_interpreter.saveContext());
	} catch (ParseException pe) {
	  out.println(pe.getMessage());
	}
      }
    });

    _commands.add("save", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: save <filename>");
	  return;
	}
	try {
	  FileOutputStream fout = new FileOutputStream(argument);
	  BufferedOutputStream bout = new BufferedOutputStream(fout, 4096);
	  ObjectOutputStream oout = new ObjectOutputStream(bout);
	  oout.writeObject(_interpreter);
	  oout.close();
	} catch (IOException ioe) {
	  out.println("I/O error: " + ioe.getMessage());
	  ioe.printStackTrace();
	}
      }
    });

    _commands.add("load", new Command() {
      public void execute(String argument) throws EvaluationException {
	if (argument.length() == 0) {
	  out.println("Usage: load <filename>");
	  return;
	}
	try {
	  FileInputStream fin = new FileInputStream(argument);
	  BufferedInputStream bin = new BufferedInputStream(fin, 4096);
	  ObjectInputStream oin = new ObjectInputStream(bin);
	  _interpreter = (Interpreter)oin.readObject();
	  _contextStack.removeAllElements();
	  oin.close();
	} catch (IOException ioe) {
	  out.println("I/O error: " + ioe.getMessage());
	  ioe.printStackTrace();
	} catch (ClassNotFoundException cnfe) {
	  out.println("Class not found: " + cnfe.getMessage());
	}
      }
    });
  }

  public java.awt.Frame createInteractionFrame() {
    return new InteractionFrame(this);
  }

  public void interrupt() { _interpreter.interrupt(); }

  public void quit() {
    if (_contextStack.size() > 1) {
      _contextStack.pop();
    } else {
      _interpreter.loadInitialContext();
      _leaveInteractor = true;
    }
  }

  protected Vector lookupCommand(String command) {
    OrderedMapIterator	iter;
    Vector		result = new Vector();

    for (iter = _commands.lowerBound(command);
	 !iter.atEnd() && ((String)iter.key()).startsWith(command);
	 iter.advance()) {
      result.addElement(iter.get());
    }

    return result;
  }

  public void execute(ApplicableValue proc, Value[] arguments) {
    try {
      synchronized (this) {
	_interpreter.invoke(proc, arguments);
      }
    } catch (EvaluationException ee) {
      ee.print(_interpreter, out);
      _contextStack.push(_interpreter.saveContext());
    }
  }

  protected synchronized
  void executeCommand(Command command, String argument) {
    try {
      command.execute(argument);
    } catch (EvaluationException ee) {
      _contextStack.pop();
      _contextStack.push(_interpreter.saveContext());
      out.println(_interpreter.getLocationString() + " " + ee.getMessage());
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable e) {
      _contextStack.pop();
      _contextStack.push(_interpreter.saveContext());
      out.print(_interpreter.getLocationString() + " " + e.getMessage());
      e.printStackTrace(out);
    }
  }

  public synchronized void executeFiles(String[] files) {
    for (int i = 0; i < files.length; ++i) {
      try {
	_interpreter.executeFile(files[i]);
      } catch (EvaluationException ee) {
	ee.print(_interpreter, out);
      } catch (FileNotFoundException fnfe) {
	out.println("File not found: \"" + files[i]);
      } catch (ParseException pe) {
	out.println(pe.getMessage());
      }
    }
  }

  public void run() throws IOException {
    _leaveInteractor = false;
    Command theCommand = null;
    String theArgument = null;

    while (! _leaveInteractor) {
	    
      _interpreter.restoreContext((Context)_contextStack.peek());
      if (_interpreter.trace) {
	out.print(System.identityHashCode(_interpreter.currentCode()) + ":");
      }
      if (_interpreter.currentCode() != null) {
	out.print(_interpreter.getLocationString());
      }
      out.print((_contextStack.size() - 1) + "> ");
      out.flush();

      String line = in.readLine();
      if (line == null) return;
      line = line.trim();

      if (line.length() > 0) {

	if (line.charAt(0) != ':') {
	  theCommand =
	    (Command)((Pair)lookupCommand("do").firstElement()).second;
	  theArgument = line;

	} else {

	  line = line.substring(1);

	  String command, argument;
	  int index, length = line.length();
	  for (index = 0; index < length; ++index) {
	    char ch = line.charAt(index);
	    if (Character.isSpaceChar(ch)) break;
	  }
	  command = line.substring(0, index);
	  argument = line.substring(index).trim();

	  Vector matchingCommands = lookupCommand(command);
	  switch (matchingCommands.size()) {
	  case 0:
	    out.println("Unknown command \"" + command + "\".");
	    continue;

	  case 1:
	    theCommand =
	      (Command)((Pair)matchingCommands.firstElement()).second;
	    theArgument = argument;
	    break;

	  default:
	    Pair top = (Pair)matchingCommands.firstElement();
	    if (top.first.equals(command)) {
	      theCommand = (Command)top.second;
	      theArgument = argument;
	      break;
	    }
	    out.println("Ambiguous command \"" + command + "\". " +
			"The following _commands has the same prefix.");
	    for (Enumeration e = matchingCommands.elements();
		 e.hasMoreElements(); ) {
	      Pair pair = (Pair)e.nextElement();
	      out.print((String)pair.first + " ");
	    }
	    out.println();
	    continue;
	  }
	}
      }

      if (theCommand == null) continue;
      executeCommand(theCommand, theArgument);
    }

    if (interactionFrame != null) interactionFrame.setVisible(false);
  }

  public static void main(String arguments[]) throws Exception {
    Interpreter interpreter = new Interpreter();
    Interactor interactor = new Interactor(interpreter, arguments[0]);
    InitialBindingLoader.execute(interpreter, interactor);
    String[] files = new String[arguments.length - 1];
    System.arraycopy(arguments, 1, files, 0, files.length);
    interactor.executeFiles(files);
    interactor.run();
    System.exit(0);
  }
}
