package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.parser.Node;
import ORG.ebiss.sara.parser.Parser;
import ORG.ebiss.sara.parser.Statement;
import ORG.ebiss.sara.parser.ParseException;

public final class Interpreter implements Serializable {
  private Code			_currentCode;
  private Frame			_currentFrame;
  private Block			_currentBlock;
  private Scope			_currentScope;
  private Value[]		_currentRegister;		
  private Scope			_globalScope;
  private boolean		_interrupted;

  public boolean	trace;
  public int[]		breakPoints;

  public Interpreter() {
    LocalNameSpace global = new LocalNameSpace();
    _globalScope = new Scope();
    _globalScope.nameSpace = global;
    _globalScope.nameSpaceToDefineVariable = global;
    _globalScope.parent = null;
    loadInitialContext();
  }

  public void interrupt() { _interrupted = true; }

  public void executeFile(String file)
       throws FileNotFoundException, ParseException, EvaluationException
  {
    String searchPath = System.getProperty("sara.path", ".");
    StringTokenizer st = new StringTokenizer(searchPath, ":");
    while (st.hasMoreTokens()) {
      File path = new File(st.nextToken(), file);
      if (path.canRead()) {
	Parser parser = new Parser(path.getAbsolutePath());
	Node node = parser.Program();
	execute(node);
      }
    }
  }

  public String getLocationString() {
    if (_currentCode != null) return _currentCode.getLocationString();
    return "<console>:0:0";
  }

  public void execute(Node node) throws EvaluationException {
    Vector labels = new Vector();
    boolean noLabelBreakIsValid = false;
    boolean noLabelContinueIsValid = false;
    for (Block block = _currentBlock; block != null; block = block.parent) {
      if (block.breakTarget != null) noLabelBreakIsValid = true;
      if (block.continueTarget != null) noLabelContinueIsValid = true;
      if (block.labels != null) {
	for (int i = block.labels.length - 1; i >= 0; --i) {
	  labels.addElement(block.labels[i]);
	}
      }
    }
    int registerSetSize =
      _currentRegister == null ? 0 : _currentRegister.length;
    Executable exec =
      Translator.translate(node, new Fence(_currentCode),
			   registerSetSize, registerSetSize,
			   labels, noLabelBreakIsValid, noLabelContinueIsValid,
			   inProcedure());
    if (_currentRegister == null ||
	_currentRegister.length < exec.registerSetSize) {
      Value[] newSet = new Value[exec.registerSetSize];
      if (_currentRegister != null) {
	System.arraycopy(newSet, 0, _currentRegister, 0,
			 _currentRegister.length);
      }
      _currentRegister = newSet;
    }
    _currentCode = exec.entryPoint;
    resume();
  }

  public Value evaluate(Node node) throws EvaluationException {
    int registerSetSize = _currentRegister == null
      ? 0 : _currentRegister.length;
    int resultSlot = registerSetSize++;
    Executable exec =
      Translator.translateExpression(node, new Fence(_currentCode), resultSlot,
				     registerSetSize, registerSetSize);
    if (_currentRegister == null ||
	_currentRegister.length < exec.registerSetSize) {
      Value[] newSet = new Value[exec.registerSetSize];
      if (_currentRegister != null) {
	System.arraycopy(newSet, 0, _currentRegister, 0,
			 _currentRegister.length);
      }
      _currentRegister = newSet;
    }
    _currentCode = exec.entryPoint;
    writeRegister(resultSlot, Value.UNSPECIFIED);
    resume();
    return readRegister(resultSlot);
  }

  public Value evaluate(ApplicableValue procedure, Value[] arguments)
       throws EvaluationException
  {
    Context context = saveContext();
    loadInitialContext();
    _currentRegister = new Value[arguments.length + 1];
    System.arraycopy(arguments, 0, _currentRegister, 0, arguments.length);
    int[] argvector = new int[arguments.length];
    for (int i = 0; i < arguments.length; ++i) argvector[i] = i;
    writeRegister(arguments.length, Value.UNSPECIFIED);
    Code code =
      procedure.apply(this, argvector, arguments.length, new Fence(null));
    _currentCode = code;
    resume();
    Value result = readRegister(arguments.length);
    restoreContext(context);
    return result;
  }

  public void invoke(ApplicableValue procedure, Value[] arguments)
       throws EvaluationException
  {
    loadInitialContext();
    _currentRegister = new Value[arguments.length + 1];
    System.arraycopy(arguments, 0, _currentRegister, 0, arguments.length);
    int[] argvector = new int[arguments.length];
    for (int i = 0; i < arguments.length; ++i) argvector[i] = i;
    writeRegister(arguments.length, Value.UNSPECIFIED);
    _currentCode = procedure.apply(this, argvector,
				   arguments.length, new Fence(null));
    resume();
  }

  public void resume() throws EvaluationException {
    if (trace || breakPoints != null) {
      genericResume();
      return;
    }
    Code pc = _currentCode;
    try {
      _interrupted = false;
      while (pc != null && !_interrupted) {
	pc = pc.execute(this);
	if (pc == null) pc = _currentFrame.whenNoMatch;
	_currentCode = pc;
      }
      if (_interrupted) {
	throw new BreakPointException("Interrupted", _currentCode);
      }
    } catch (FenceException fe) {
      _currentCode = fe.resumePoint;
    } catch (BreakPointException bpe) {
      _currentCode = bpe.resumePoint;
      throw bpe;
    } catch (EvaluationException ee) {
      throw ee;
    }
  }

  private void genericResume() throws EvaluationException {
    try {
      _interrupted = false;
      do {
	if (trace) System.out.println(_currentCode.toString());
	if (_interrupted) {
	  throw new BreakPointException("Interrupted", _currentCode);
	}
	Code next = _currentCode.execute(this);
	_currentCode = next == null ? _currentFrame.whenNoMatch : next;
	if (breakPoints != null) {
	  for (int i = breakPoints.length - 1; i >= 0; --i) {
	    if (breakPoints[i] == _currentCode.hashCode()) {
	      throw new BreakPointException("break point #" + i,
					    _currentCode);
	    }
	  }
	}
      } while (_currentCode != null);
    } catch (FenceException fe) {
      _currentCode = fe.resumePoint;
    } catch (BreakPointException bpe) {
      _currentCode = bpe.resumePoint;
      throw bpe;
    } catch (EvaluationException ee) {
      throw ee;
    }
  }

  public void istep() throws EvaluationException {
    try {
      if (trace) System.out.println(_currentCode.toString());
      Code next = _currentCode.execute(this);
      _currentCode = next == null ? _currentFrame.whenNoMatch : next;
    } catch (FenceException fe) {
      _currentCode = fe.resumePoint;
    } catch (BreakPointException bpe) {
      _currentCode = bpe.resumePoint;
      throw bpe;
    } catch (EvaluationException ee) {
      throw ee;
    }
  }

  public String dumpCallStack() {
    StringBuffer buffer = new StringBuffer();
    if (inProcedure()) {
      Code code = _currentCode;
      for (Frame frame = _currentFrame; frame != null; frame = frame.parent) {
	buffer.append(code.getLocationString() + " ");
	PrintingContext context = new PrintingContext();
	if (frame.arguments != null) {
	  buffer.append('(');
	  for (int i = 0; i < frame.arguments.length; ++i) {
	    if (i > 0) buffer.append(", ");
	    Value arg = frame.register[frame.arguments[i]];
	    buffer.append(arg.getPrintableRepresentation(context));
	  }
	  buffer.append(')');
	}
	buffer.append('\n');
	code = frame.returnPoint;
      }
    }
    return buffer.toString();
  }

  public void finishProcedure(Value resultValue) {
    _currentCode = popFrameAndReturn(resultValue);
  }

  public NameSpace getGlobalNameSpace() { return _globalScope.nameSpace; }

  public Code currentCode() { return _currentCode; }

  public Node currentNode() {
    return (_currentCode == null || _currentCode.node == null)
      ? null : _currentCode.node;
  }

  public void setCurrentCode(Code code) { _currentCode = code; }

  public void resetNameSpace() { _currentScope = _globalScope; }

  public Scope getScope() { return _currentScope; }

  public void setScope(Scope scope) { _currentScope = scope; }

  public int arity() { return _currentFrame.arguments.length; }

  public int[] getArgumentsVector() { return _currentFrame.arguments; }

  public void createRegisterSet(int size) {
    if (_currentRegister == null || size > _currentRegister.length) {
      _currentRegister = new Value[size];
    }
  }

  public void loadInitialContext() {
    _currentCode = null;
    _currentFrame = new Frame();
    _currentBlock = new Block();
    _currentRegister = null;
    _currentScope = _globalScope;
  }

  public Context saveContext() {
    Context context = new Context();
    context.code = _currentCode;
    context.frame = _currentFrame;
    context.block = _currentBlock;
    context.scope = _currentScope;
    context.register = _currentRegister;
    return context;
  }

  public void restoreContext(Context context) {
    _currentCode = context.code;
    _currentFrame = context.frame;
    _currentBlock = context.block;
    _currentScope = context.scope;
    _currentRegister = context.register;
  }

  public void pushFrame(int[] arguments, int resultSlot, Code returnPoint,
			Code whenNoMatch) {
    Frame newFrame = new Frame();
    newFrame.parent = _currentFrame;
    _currentFrame = newFrame;
    newFrame.block = _currentBlock;
    _currentBlock = new Block();
    newFrame.scope = _currentScope;
    _currentScope = _globalScope;
    newFrame.register = _currentRegister;
    _currentRegister = null;
    newFrame.arguments = arguments;
    newFrame.resultSlot = resultSlot;
    newFrame.returnPoint = returnPoint;
    newFrame.whenNoMatch = whenNoMatch;
  }

  public void pushFrame(int[] arguments, Code returnPoint, Code whenNoMatch) {
    Frame newFrame = new Frame();
    newFrame.parent = _currentFrame;
    _currentFrame = newFrame;
    newFrame.block = _currentBlock;
    _currentBlock = new Block();
    newFrame.scope = _currentScope;
    _currentScope = _globalScope;
    newFrame.register = _currentRegister;
    _currentRegister = null;
    newFrame.arguments = arguments;
    newFrame.resultSlot = -1;
    newFrame.returnPoint = returnPoint;
    newFrame.whenNoMatch = whenNoMatch;
  }

  public Code popFrame() {
    Code returnPoint = _currentFrame.returnPoint;
    _currentRegister = _currentFrame.register;
    _currentScope = _currentFrame.scope;
    _currentBlock = _currentFrame.block;
    _currentFrame = _currentFrame.parent;
    return returnPoint;
  }

  public Code popFrameAndReturn(Value resultValue) {
    if (_currentFrame.resultSlot >= 0) {
      _currentFrame.register[_currentFrame.resultSlot] = resultValue;
    }
    Code returnPoint = _currentFrame.returnPoint;
    _currentRegister = _currentFrame.register;
    _currentScope = _currentFrame.scope;
    _currentBlock = _currentFrame.block;
    _currentFrame = _currentFrame.parent;
    return returnPoint;
  }

  public boolean inProcedure() { return _currentFrame.parent != null; }

  public Code whenNoMatch() { return _currentFrame.whenNoMatch; }

  public Code getReturnPoint() { return _currentFrame.returnPoint; }

  public Value readArgument(int index) {
    return _currentFrame.register[_currentFrame.arguments[index]];
  }

  public Value readRegister(int index) {
    return _currentRegister[index];
  }

  public void writeRegister(int index, Value value) {
    _currentRegister[index] = value;
  }

  public void pushNameSpace(NameSpace nameSpace) {
    Scope scope = new Scope();
    scope.parent = _currentScope;
    scope.nameSpace = nameSpace;
    scope.nameSpaceToDefineVariable = nameSpace;
    _currentScope = scope;
  }

  public void pushReadOnlyNameSpace(NameSpace nameSpace) {
    Scope scope = new Scope();
    scope.parent = _currentScope;
    scope.nameSpace = nameSpace;
    scope.nameSpaceToDefineVariable = scope.parent.nameSpace;
    _currentScope = scope;
  }

  public void popNameSpace() {
    _currentScope = _currentScope.parent;
  }

  public void pushBreakTarget(Code breakTarget) {
    Block block = new Block();
    block.parent = _currentBlock;
    block.scope = _currentScope;
    block.breakTarget = breakTarget;
    _currentBlock = block;
  }

  public void pushBreakContinueTargets(Code breakTarget, Code continueTarget) {
    Block block = new Block();
    block.parent = _currentBlock;
    block.scope = _currentScope;
    block.breakTarget = breakTarget;
    block.continueTarget = continueTarget;
    _currentBlock = block;
  }

  public void pushLabeledTargets(String[] labels,
				 Code breakTarget, Code continueTarget)
  {
    Block block = new Block();
    block.parent = _currentBlock;
    block.scope = _currentScope;
    block.labels = labels;
    block.breakTarget = breakTarget;
    block.continueTarget = continueTarget;
    _currentBlock = block;
  }

  public void popTarget() {
    _currentScope = _currentBlock.scope;
    _currentBlock = _currentBlock.parent;
  }

  public Code doBreak() {
    for (Block block = _currentBlock; block != null; block = block.parent) {
      if (block.breakTarget != null) {
	_currentScope = block.scope;
	return block.breakTarget;
      }
    }
    return null;
  }

  public Code doBreak(String label) {
    Block block = findLabel(label);
    if (block == null) return null;
    _currentScope = block.scope;
    return block.breakTarget;
  }

  public Code doContinue() {
    for (Block block = _currentBlock; block != null; block = block.parent) {
      if (block.continueTarget != null) {
	_currentScope = block.scope;
	return block.continueTarget;
      }
    }
    return null;
  }

  public Code doContinue(String label) {
    Block block = findLabel(label);
    if (block == null) return null;
    _currentScope = block.scope;
    return block.continueTarget;
  }

  private Block findLabel(String label) {
    for (Block block = _currentBlock; block != null; block = block.parent) {
      if (block.labels == null) continue;
      for (int i = block.labels.length - 1; i >= 0; --i) {
	if (block.labels[i].equals(block.labels)) return block;
      }
    }
    return null;
  }

  public Value readVariable(String name) {
    for (Scope scope = _currentScope; scope != null; scope = scope.parent) {
      Value result = scope.nameSpace.read(name);
      if (result != null) return result;
    }
    return null;
  }

  public boolean writeVariable(String name, Value value)
       throws InvalidOperationException
  {
    for (Scope scope = _currentScope; scope != null; scope = scope.parent) {
      if (scope.nameSpace.write(name, value)) return true;
    }
    return false;
  }

  public boolean defineVariable(String name, Value value)
       throws InvalidOperationException
  {
    if (_currentScope.nameSpaceToDefineVariable.read(name) != null) {
      return false;
    }
    _currentScope.nameSpaceToDefineVariable.define(name, value);
    return true;
  }

  public boolean isDefinedVariable(String name) {
    for (Scope scope = _currentScope; scope != null; scope = scope.parent) {
      if (scope.nameSpace.contains(name)) return true;
    }
    return false;
  }
}
