package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.parser.Node;

public final
class PatternTranslator extends ORG.ebiss.sara.parser.PartialNodeVisitor {
  protected Pattern	result;

  public Pattern translate(Node pattern) {
    try {
      pattern.accept(this);
    } catch (Exception e) {
      throw new RuntimeException(e.getMessage());
    }
    return result;
  }

  public void visit(ORG.ebiss.sara.parser.LiteralPattern x)
       throws Exception
  {
    LiteralPattern pattern = new LiteralPattern();
    pattern.value = Translator.translateLiteral(x.value);
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.VariablePattern x)
       throws Exception
  {
    VariablePattern pattern = new VariablePattern();
    pattern.name = x.name;
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.ListPattern x)
       throws Exception
  {
    ListPattern pattern = new ListPattern();
    pattern.name = x.name;
    pattern.elements = new Pattern[x.elements.size()];
    for (int i = pattern.elements.length - 1; i >= 0; --i) {
      ((Node)x.elements.elementAt(i)).accept(this);
      pattern.elements[i] = result;
    }
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.VectorPattern x)
       throws Exception
  {
    VectorPattern pattern = new VectorPattern();
    pattern.name = x.name;
    pattern.elements = new Pattern[x.elements.size()];
    for (int i = pattern.elements.length - 1; i >= 0; --i) {
      ((Node)x.elements.elementAt(i)).accept(this);
      pattern.elements[i] = result;
    }
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.ObjectPattern x)
       throws Exception
  {
    ObjectPattern pattern = new ObjectPattern();
    pattern.name = x.name;
    pattern.members = new Pattern[x.members.size()];
    for (int i = pattern.members.length - 1; i >= 0; --i) {
      ((Node)x.members.elementAt(i)).accept(this);
      pattern.members[i] = result;
    }
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.InstanceVariablePattern x)
       throws Exception
  {
    InstanceVariablePattern pattern = new InstanceVariablePattern();
    pattern.name = x.name;
    x.value.accept(this);
    pattern.valuePattern = result;
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.MethodDefinitionPattern x)
       throws Exception
  {
    MethodDefinitionPattern pattern = new MethodDefinitionPattern();
    pattern.name = x.name;
    pattern.arity = x.arity;
    result = pattern;
  }

  public void visit(ORG.ebiss.sara.parser.CatchAllPattern x)
       throws Exception
  {
    result = new CatchAllPattern();
  }
}

final class LiteralPattern implements Pattern {
  public Value	value;

  public final boolean match(Interpreter interpreter, Value matchedValue) {
    return value.equals(matchedValue);
  }
}

final class CatchAllPattern implements Pattern {
  public final boolean match(Interpreter interpreter, Value matchedValue) {
    return true;
  }
}

final class VariablePattern implements Pattern {
  public String	name;

  public final boolean match(Interpreter interpreter, Value matchedValue) {
    // Value value = interpreter.getScope().nameSpace.read(name);
    // if (value != null) return value.equals(matchedValue);
    try {
      interpreter.defineVariable(name, matchedValue);
    } catch (InvalidOperationException usoe) {
      usoe.printStackTrace();
      throw new RuntimeException(usoe.getMessage());
    }
    return true;
  }
}

class SequencePattern implements Pattern {
  public Pattern[]	elements;
  public String		name;

  public boolean match(Interpreter interpreter, Value matchedValue) {
    Sequence list = (Sequence)((ContainerValue)matchedValue).getJavaValue();
    if (elements.length != list.size()) return false;
    Enumeration e = list.elements();
    for (int i = 0; i < elements.length; ++i) {
      if (! elements[i].match(interpreter, (Value)e.nextElement())) {
	return false;
      }
    }
    if (name != null) {
      try {
	interpreter.defineVariable(name, matchedValue);
      } catch (InvalidOperationException uoe) {
	uoe.printStackTrace();
	throw new RuntimeException();
      }
    }
    return true;
  }
}

final class ListPattern extends SequencePattern {
  public boolean match(Interpreter interpreter, Value matchedValue) {
    if (! (matchedValue instanceof ListValue)) return false;
    return super.match(interpreter, matchedValue);
  }
}

final class VectorPattern extends SequencePattern {
  public boolean match(Interpreter interpreter, Value matchedValue) {
    if (! (matchedValue instanceof VectorValue)) return false;
    return super.match(interpreter, matchedValue);
  }
}

final class ObjectPattern implements Pattern {
  public Pattern[]	members;
  public String		name;

  public final boolean match(Interpreter interpreter, Value matchedValue) {
    for (int i = 0; i < members.length; ++i) {
      if (! members[i].match(interpreter, matchedValue)) return false;
    }
    if (name != null) {
      try {
	interpreter.defineVariable(name, matchedValue);
      } catch (InvalidOperationException uoe) {
	uoe.printStackTrace();
	throw new RuntimeException();
      }
    }
    return true;
  }
}

final class InstanceVariablePattern implements Pattern {
  public String		name;
  public Pattern	valuePattern;

  public final boolean match(Interpreter interpreter, Value matchedValue) {
    Value value = matchedValue.read(name);
    if (value == null) return false;
    return valuePattern.match(interpreter, value);
  }
}

final class MethodDefinitionPattern implements Pattern {
  public String		name;
  public int		arity;

  public final boolean match(Interpreter interpreter, Value matchedValue) {
    Value value = matchedValue.read(name);
    if (value == null) return false;
    return ((ApplicableValue)value).checkArity(arity);
  }
}
