package ORG.ebiss.sara.server;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.interpreter.*;
import ORG.ebiss.sara.interpreter.InvalidOperationException;

public class DataConverter {

  public static
  UIAtomic encodeInt(int type, String label, int value) {
    UIAtomic result = new UIAtomic();
    result.type = type;
    result.label = label;
    result.value = new Integer(value);
    return result;
  }

  public static
  UIAtomic encodeBoolean(int type, String label, boolean value) {
    UIAtomic result = new UIAtomic();
    result.type = type;
    result.label = label;
    result.value = value ? Boolean.TRUE : Boolean.FALSE;
    return result;
  }

  public static UIAtomic encodeSymbol(int type, String label,
						  SymbolValue value)
  {
    UIAtomic result = new UIAtomic();
    result.type = type;
    result.label = label;
    result.value = value.name();
    return result;
  }

  public static
  UIObject encodeDouble(int type, String label, double value) {
    UIAtomic result = new UIAtomic();
    result.type = type;
    result.label = label;
    result.value = new Double(value);
    return result;
  }

  public static
  UIObject encodeString(int type, String label, String value) {
    UIAtomic result = new UIAtomic();
    result.type = type;
    result.label = label;
    result.value = value;
    return result;
  }

  public static UIObject encodeChoice(String label,
						  ContainerValue contents)
  {
    UIChoice result = new UIChoice();
    result.description = label;
    result.index = 0;
    result.labels = new Vector(contents.size());
    result.values = new Vector(contents.size());
    for (Enumeration e = contents.elements(); e.hasMoreElements(); ) {
      Value item = (Value)e.nextElement();
      if (item instanceof ContainerValue) {
	result.labels.addElement(encode("", ((ContainerValue)item).at(0)));
	result.values.addElement(encode("", ((ContainerValue)item).at(1)));
      } else {
	result.labels.addElement(encode("", item));
	result.values.addElement(encode("", item));
      }
    }
    return result;
  }

  public static UIObject encodeCheckBox(String label,
						    ContainerValue contents)
  {
    try {
      UICheckBox result = new UICheckBox();
      result.description = label;
      result.labels = new Vector(contents.size());
      result.checks = new Vector(contents.size());
      for (Enumeration e = contents.elements(); e.hasMoreElements(); ) {
	Value item = (Value)e.nextElement();
	if (item instanceof ContainerValue) {
	  result.labels.addElement(encode("", ((ContainerValue)item).at(0)));
	  result.checks.addElement(((ContainerValue)item).at(1).booleanValue()
				   ? Boolean.TRUE : Boolean.FALSE);
	} else {
	  result.labels.addElement(encode("", item));
	  result.checks.addElement(Boolean.FALSE);
	}
      }
      return result;
    } catch (CoerceFailedException cfe) {
      cfe.printStackTrace();
      throw new RuntimeException();
    }
  }

  public static
  UIObject encode(String label, Value value) {
    try {
      if (value == Value.NULL) return null;

      if (value instanceof IntegerValue) {
	return encodeInt(UIObject.IS_A_INTEGER,
			 label, value.intValue());
      }

      if (value instanceof DoubleValue) {
	return encodeDouble(UIObject.IS_A_FLOAT,
			    label, value.doubleValue());
      }

      if (value instanceof StringValue) {
	return encodeString(UIObject.IS_A_SHORT_STRING,
			    label, value.stringValue());
      }

      if (value instanceof BooleanValue) {
	return encodeBoolean(UIObject.IS_A_BOOLEAN,
			     label, value.booleanValue());
      }

      if (value instanceof SymbolValue) {
	return encodeSymbol(UIObject.IS_A_SYMBOL,
			    label, (SymbolValue)value);
      }

      if (value instanceof ContainerValue) {
	UIContainer result = new UIContainer();
	ContainerValue container = (ContainerValue)value;
	result.description = label;
	result.className = container.getClass().getName();
	result.values = new Vector(container.size());
	for (Enumeration e = container.elements(); e.hasMoreElements(); ) {
	  result.values.addElement(encode("", (Value)e.nextElement()));
	}
	return result;
      }

      if (value instanceof JavaValue) {
	Object o = ((JavaValue)value).getJavaValue();
	if (o instanceof UIObject) return (UIObject)o;
      }

      if (value instanceof SpdValue) {
	UIRecord result = new UIRecord();
	result.description = label;
	result.values = new Hashtable();
	SpdValue spd = (SpdValue)value;
	for (Enumeration e = spd.elements(); e.hasMoreElements(); ) {
	  Slot slot = (Slot)e.nextElement();
	  String name = slot.name();
	  if (slot instanceof ObservableSlot) {
	    label = ((ObservableSlot)slot).label();
	  } else {
	    label = name;
	  }
	  result.values.put(name, encode(label, spd.read(name)));
	}
	return result;
      }

      if (value == Value.NULL) {
	throw new RuntimeException("Can\'t encode the null value");
      }
      if (value == Value.UNSPECIFIED) {
	throw new RuntimeException("Can\'t encode the <unspecified> value");
      }

      throw new RuntimeException(value.getClass().getName());
    } catch (CoerceFailedException cfe) {
      cfe.printStackTrace();
      throw new RuntimeException();
    }
  }

  public static Value decode(UIObject object) {
    switch (object.getType()) {
    case UIObject.IS_A_INTEGER:
      {
	Integer value = (Integer)((UIAtomic)object).value;
	return new IntegerValue(value.intValue());
      }

    case UIObject.IS_A_FLOAT:
      {
	Double value = (Double)((UIAtomic)object).value;
	return new DoubleValue(value.doubleValue());
      }

    case UIObject.IS_A_SHORT_STRING:
      {
	String value = (String)((UIAtomic)object).value;
	return new StringValue(value);
      }

    case UIObject.IS_A_BOOLEAN:
      {
	Boolean value = (Boolean)((UIAtomic)object).value;
	return value.booleanValue() ? BooleanValue.TRUE : BooleanValue.FALSE;
      }

    case UIObject.IS_A_SYMBOL:
      {
	String name = (String)((UIAtomic)object).value;
	return SymbolValue.intern(name);
      }

    case UIObject.IS_A_LONG_STRING:
    case UIObject.IS_A_LABEL:
    case UIObject.IS_A_TEXT:
    case UIObject.IS_A_CHOICE:
    case UIObject.IS_A_CHECKBOX:
    case UIObject.IS_A_INTEGER_PARAM:
    case UIObject.IS_A_FLOAT_PARAM:
    case UIObject.IS_A_SLIDER:      
      return new JavaValue(object);
      
    case UIObject.IS_COMPOSITE:
      if (object instanceof UIRecord) {
	try {
	  SpdValue result = new SpdValue();
	  Hashtable table = ((UIRecord)object).values;
	  for (Enumeration e = table.keys(); e.hasMoreElements(); ) {
	    String key = (String)e.nextElement();
	    result.
	      defineVariable(SpdValue.noModifiers, key,
			     decode((UIObject)table.get(key)));
	  }
	  return result;
	} catch (InvalidOperationException uoe) {
	  uoe.printStackTrace();
	  throw new RuntimeException();
	}
      }

      try {
	UIContainer container = (UIContainer)object;
	ContainerValue result =
	  (ContainerValue)Class.forName(container.className).newInstance();
	for (Enumeration e = container.values.elements();
	     e.hasMoreElements(); ) {
	  result.add(decode((UIObject)e.nextElement()));
	}
	return result;
      } catch (ClassNotFoundException cnfe) {
	cnfe.printStackTrace();
	throw new RuntimeException();
      } catch (InstantiationException ie) {
	ie.printStackTrace();
	throw new RuntimeException();
      } catch (IllegalAccessException iae) {
	iae.printStackTrace();
	throw new RuntimeException();
      }

    default:
      throw new RuntimeException();
    }
  }
}
