package ORG.ebiss.sara.server;

import java.io.*;
import java.net.*;
import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.interpreter.*;
import ORG.ebiss.sara.parser.Node;
import ORG.ebiss.sara.parser.Parser;
import ORG.ebiss.sara.parser.ParseException;
import ORG.ebiss.sara.parser.LocalizedCharStream;

public class Server extends Interactor {

  public final static int	SERVER_PORT = 8888;
  public Timer			timer;

  private boolean		_pause;
  private Listener		_listener = new Listener();
  private Queue			_commandQueue = new Queue();
  private DList			_connections = new DList();

  public Server(Interpreter interpreter, String consoleServerPath)
       throws IOException, NumberFormatException
  {
    super(interpreter, consoleServerPath);

    timer = new Timer(this);

    _listener.start();

    _commands.add("freeze", new Command() {
      public void execute(String argument) {
	if (argument.length() == 0) {
	  out.println("freeze = " + (_pause ? "on" : "off"));
	  return;
	}
	if (argument.equals("on")) {
	  pause();
	  return;
	}
	if (argument.equals("off")) {
	  resume();
	  return;
	}
	out.println("Usage: freeze [(on|off)]");
      }
    });

    _commands.add("timer", new Command() {
      public void execute(String argument) {
	if (argument.length() == 0) {
	  out.println("Current timer setting: " +
		      (timer.isRunning() ? "on " : "off ") +
		      "period = " + timer.getPeriod());
	  return;
	}
	StringTokenizer tokenizer = new StringTokenizer(argument);
	while (tokenizer.hasMoreTokens()) {
	  String token = tokenizer.nextToken();
	  if      (token.equals("on"))  timer.setRunning(true);
	  else if (token.equals("off")) timer.setRunning(false);
	  else try {
	    timer.setPeriod(Integer.parseInt(token));
	  } catch (NumberFormatException nfe) {
	    out.println("Usage: timer (on | off | <period>)*");
	  }
	}
      }
    });

    _commands.add("connections", new Command() {
      public void execute(String argument) {
	int i = 1;
	for (Enumeration e = _connections.elements();
	     e.hasMoreElements(); ++i) {
	  ServerPort port = (ServerPort)e.nextElement();
	  Socket socket = port.socket();
	  InetAddress address = socket.getInetAddress();
	  out.print(i);
	  out.print(": " + address.getHostName() + "#" + socket.getPort());
	  for (Enumeration f = port.subjects(); f.hasMoreElements(); ) {
	    out.print(", ");
	    out.print(f.nextElement());
	  }
	  out.println();
	}
      }
    });
  }

  public synchronized void pause() {
    if (_pause) return;
    _pause = true;
    timer.setRunning(false);
  }

  public synchronized void resume() {
    if (!_pause) return;
    _pause = false;
    while (! _commandQueue.isEmpty()) {
      ServerPort port = (ServerPort)_commandQueue.pop();
      ServerCommand command = (ServerCommand)_commandQueue.pop();
      command.execute(this, _interpreter, port);
    }
    timer.setRunning(true);
  }

  protected
  void executeCommand(Command command, String argument) {
    try {
      synchronized(this) {
	command.execute(argument);
      }
    } catch (EvaluationException ee) {
      _contextStack.pop();
      _contextStack.push(_interpreter.saveContext());
      out.println(_interpreter.getLocationString() + " " + ee.getMessage());
      // pause();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable e) {
      _contextStack.pop();
      _contextStack.push(_interpreter.saveContext());
      out.print(_interpreter.getLocationString() + " " + e.getMessage());
      e.printStackTrace(out);
      // pause();
    }
  }

  public void registerPort(ServerPort port) {
    _connections.add(port);
  }

  public void unregisterPort(ServerPort port) {
    _connections.remove(port);
  }

  public void accept(ServerPort receiver, ServerCommand command) {
    synchronized (this) {
      if (_pause) {
	_commandQueue.push(receiver);
	_commandQueue.push(command);
      } else {
	command.execute(this, _interpreter, receiver);
      }
    }
    timer.restart();
  }

  public void sendChangeList(Container changeList) {
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.sendChangeList(changeList);
    }
  }

  public void sendParameterList(String id, String actionName,
				final Port port, final int key) {
    Value postOffice = _interpreter.readVariable("thePostOffice");
    if (postOffice == null) throw new RuntimeException();
    ApplicableValue getParameterList =
      (ApplicableValue)postOffice.read("passParameterList");
    if (getParameterList == null) throw new RuntimeException();
    Value[] arguments = new Value[3];
    arguments[0] = SymbolValue.intern(id);
    arguments[1] = new StringValue(actionName);
    arguments[2] = new ApplicableValue() {
      public Code apply(Interpreter interpreter, int[] arguments,
			int resultSlot, Code returnPoint) {
	Value result = interpreter.readRegister(arguments[0]);
	UIObject object = DataConverter.encode("", result);
	ReplyToRequest reply = new ReplyToRequest();
	reply.replyTo = key;
	reply.result = object;
	port.send(reply);
	return returnPoint;
      }

      public boolean checkArity(int arity) { return arity == 1; }
    };
    execute(getParameterList, arguments);
  }

  public void invokeAction(String id, String actionName,
			   UIObject parameterList,
			   final Port port, final int key) {
    Value value = null;
    Value postOffice = _interpreter.readVariable("thePostOffice");
    if (postOffice == null) throw new RuntimeException();
    ApplicableValue invokeAction =
      (ApplicableValue) postOffice.read("invokeAction");
    if (invokeAction == null) throw new RuntimeException();
    Value[] arguments = new Value[4];
    arguments[0] = SymbolValue.intern(id);
    arguments[1] = SymbolValue.intern(actionName);
    arguments[2] = DataConverter.decode(parameterList);
    arguments[3] = new ApplicableValue() {
      public Code apply(Interpreter interpreter, int[] arguments,
			int resultSlot, Code returnPoint) {
	boolean result = false;
	try {
	  result = interpreter.readRegister(arguments[0]).booleanValue();
	} catch (Exception e) {
	  e.printStackTrace();
	}
	ReplyToRequest reply = new ReplyToRequest();
	reply.replyTo = key;
	reply.result = result ? Boolean.TRUE : Boolean.FALSE;
	port.send(reply);
	return returnPoint;
      }

      public boolean checkArity(int arity) { return arity == 1; }
    };
    execute(invokeAction, arguments);
  }

  public void sendUIState(String id, ServerPort port) {
    Value value = null;
    Value postOffice = _interpreter.readVariable("thePostOffice");
    if (postOffice == null) throw new RuntimeException();
    ApplicableValue sendUIState =
      (ApplicableValue)postOffice.read("sendUIState");
    if (sendUIState == null) throw new RuntimeException();
    Value[] arguments = new Value[2];
    arguments[0] = SymbolValue.intern(id);
    arguments[1] = new JavaValue(port);
    execute(sendUIState, arguments);
  }

  public void sendAgentState(String id, ServerPort port) {
    Value value = null;
    Value postOffice = _interpreter.readVariable("thePostOffice");
    if (postOffice == null) throw new RuntimeException();
    ApplicableValue sendAgentState =
      (ApplicableValue)postOffice.read("sendAgentState");
    if (sendAgentState == null) throw new RuntimeException();
    Value[] arguments = new Value[2];
    arguments[0] = SymbolValue.intern(id);
    arguments[1] = new JavaValue(port);
    execute(sendAgentState, arguments);
  }

  public void show(String id, Value value) {
    UIObject object = DataConverter.encode("", value);
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.show(id, object);
    }
  }

  public void sendInteractionRequest(ServerPort port, String id,
				     int key, Value parameter) {
    InteractionRequestCommand command = new InteractionRequestCommand();
    command.id = id;
    command.key = key;
    command.parameter =
      (CompositeUIObject) DataConverter.encode("", parameter);
    port.send(id, command);
  }

  public void sendInteractionRequest(String id, int key, Value parameter) {
    InteractionRequestCommand command = new InteractionRequestCommand();
    command.id = id;
    command.key = key;
    command.parameter =
      (CompositeUIObject)DataConverter.encode("", parameter);
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.send(id, command);
    }
  }

  public void retractInteractionRequest(String id, int key) {
    RetractInteractRequestCommand command =
      new RetractInteractRequestCommand();
    command.id = id;
    command.key = key;
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.send(id, command);
    }
  }

  public void notifyAgentAuxiliaryCreation(String agentId,
					   String type, String auxId) {
    CreateAgentAuxCommand command = new CreateAgentAuxCommand();
    command.agentId = agentId;
    command.auxId = auxId;
    command.type = type;
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.send(agentId, command);
    }
  }

  public void notifyAgentAuxiliaryCreation(ServerPort port, String agentId,
					   String type, String auxId) {
    CreateAgentAuxCommand command = new CreateAgentAuxCommand();
    command.agentId = agentId;
    command.auxId = auxId;
    command.type = type;
    port.send(agentId, command);
  }

  public void notifyAgentAuxiliaryDestruction(String agentId,
					      String type, String auxId) {
    DeleteAgentAuxCommand command = new DeleteAgentAuxCommand();
    command.agentId = agentId;
    command.auxId = auxId;
    command.type = type;
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.send(agentId, command);
    }
  }

    public void sendExtendedInformation(String agentId,
				      Value type, Value uiComponent) {
    NotifyExtendedInfoCommand command = new NotifyExtendedInfoCommand();
    command.agentId = agentId;
    String[] typev = new String[((ContainerValue)type).size()];
    int i = 0;
    for (Enumeration e = ((ContainerValue)type).elements();
	 e.hasMoreElements(); ) {
      typev[i++] = e.nextElement().toString();
    }
    command.type = typev;
    command.info = DataConverter.encode("", uiComponent);
    for (Enumeration e = _connections.elements(); e.hasMoreElements(); ) {
      ServerPort port = (ServerPort)e.nextElement();
      port.send(agentId, command);
    }
  }

  public void sendExtendedInformation(ServerPort port, String agentId,
				      Value type, Value uiComponent) {
    NotifyExtendedInfoCommand command = new NotifyExtendedInfoCommand();
    command.agentId = agentId;
    String[] typev = new String[((ContainerValue)type).size()];
    int i = 0;
    for (Enumeration e = ((ContainerValue)type).elements();
	 e.hasMoreElements(); ) {
      typev[i++] = e.nextElement().toString();
    }
    command.type = typev;
    command.info = DataConverter.encode("", uiComponent);
    port.send(agentId, command);
  }

  public void acceptReplyInteraction(int key,
				     UIObject parameter,
				     final Port port, final int replyKey)
  {
    ReplyToRequest reply = new ReplyToRequest();
    reply.replyTo = replyKey;
    reply.result = Boolean.TRUE;
    port.send(reply);

    Value value = null;
    Value postOffice = _interpreter.readVariable("thePostOffice");
    if (postOffice == null) throw new RuntimeException();
    ApplicableValue acceptReplyInteraction =
      (ApplicableValue)postOffice.read("acceptReplyInteraction");
    if (acceptReplyInteraction == null) throw new RuntimeException();
    Value[] arguments = new Value[3];
    arguments[0] = new IntegerValue(key);
    arguments[1] = DataConverter.decode(parameter);
    arguments[2] = new ApplicableValue() {
      public Code apply(Interpreter interpreter, int[] arguments,
			int resultSlot, Code returnPoint) {
	/*
	boolean result = false;
	try {
	  result = interpreter.readRegister(arguments[0]).booleanValue();
	} catch (Exception e) {
	  e.printStackTrace();
	}
	*/
	return returnPoint;
      }

      public boolean checkArity(int arity) { return arity == 1; }
    };
    execute(acceptReplyInteraction, arguments);
  }

  public Vector connections(){
    int i = 1;
    Vector result = new Vector();
    for (Enumeration e = _connections.elements();e.hasMoreElements(); ++i) {
      ServerPort port = (ServerPort)e.nextElement();
      Socket socket = port.socket();
      InetAddress address = socket.getInetAddress();
      Vector list = new Vector();
      for (Enumeration f = port.subjects(); f.hasMoreElements();){
	list.addElement(f.nextElement());
      }
      Vector v = new Vector();
      v.addElement(address.getHostName());
      v.addElement(list);
      result.addElement(v);
    }
    return result;
  }

  public static class InteractionRequestCommand extends ClientCommand {
    String				id;
    int					key;
    CompositeUIObject	parameter;

    public void writeObject(OldObjectOutputStream out) throws IOException {
      super.writeObject(out);
      out.writeUTF(id);
      out.writeInt(key);
      out.writeObject(parameter);
    }

    public void readObject(OldObjectInputStream in)
	 throws IOException, ClassNotFoundException, IllegalAccessException,
		InstantiationException
    {
      super.readObject(in);
      id = in.readUTF();
      key = in.readInt();
      parameter = (CompositeUIObject)in.readObject();
    }

    public void execute() {
      ((SimValueAdapter)client.getAdapter(id)).
	acceptInteractionRequest(key,
				 (CompositeUIObject)parameter);
    }
  }

  public static class RetractInteractRequestCommand extends ClientCommand {
    public String	id;
    public int		key;

    public void writeObject(OldObjectOutputStream out) throws IOException {
      super.writeObject(out);
      out.writeUTF(id);
      out.writeInt(key);
    }

    public void readObject(OldObjectInputStream in)
	 throws IOException, ClassNotFoundException, IllegalAccessException,
		InstantiationException
    {
      super.readObject(in);
      id = in.readUTF();
      key = in.readInt();
    }

    public void execute() {
      ((SimValueAdapter)client.getAdapter(id)).retractInteractionRequest(key);
    }
  }

  public static Change createChangeRecord(Value value) {
    Change result = new Change();
    try {
      result.id = value.read("id").stringValue();
      Value typeValue = value.read("type");
      if (typeValue == null) {
	result.type = Change.DUMP;
      } else {
	String type = typeValue.stringValue();
	if (type.equals("add")) {
	  result.type = Change.ADD;
	} else if (type.equals("remove")) {
	  result.type = Change.REMOVE;
	} else if (type.equals("modify")) {
	  result.type = Change.MODIFY;
	} else {
	  throw new RuntimeException();
	}
      }
      result.slotName = value.read("name").stringValue();
      result.label = value.read("label").stringValue();
      result.visible = value.read("visible").booleanValue();
      result.action = value.read("action").booleanValue();
      Value v = value.read("value");
      if (v != null) {
	result.value = DataConverter.encode(result.label, v);
      }
      return result;
    } catch (CoerceFailedException cfe) {
      cfe.printStackTrace();
      throw new RuntimeException();
    }
  }

  class Listener extends Thread {
    Listener() {
      setName(getClass().getName());
      setDaemon(true);
    }

    public void run() {
      try {
	ServerSocket socket = new ServerSocket(SERVER_PORT);
	for (;;) {
	  ServerPort port = new ServerPort(Server.this, socket.accept());
	}
      } catch (IOException e) {
	e.printStackTrace(out);
	throw new RuntimeException();
      }
    }
  }

  public static void main(String arguments[]) throws Exception {
    Interpreter interpreter = new Interpreter();
    Server server = new Server(interpreter, arguments[0]);
    server.pause();
    ServerBindingLoader.execute(interpreter, server);
    String[] files = new String[arguments.length - 1];
    System.arraycopy(arguments, 1, files, 0, files.length);
    server.executeFiles(files);
    server.resume();
    server.run();
    System.exit(0);
  }




}
